#include <windows.h>
#include <string.h>
#include <mmsystem.h>

#ifdef __cplusplus
extern "C" {
#endif

#define VERSION "1.11"

#define AUTHINFO \
"Media Control Interface (MCI) DLL for mIRC, v" VERSION " (c) 1999 DragonZap, dragonzap1@aol.com"

DWORD ControlDataType(DWORD MixerType);
DWORD GetLineID(const char *idStr,DWORD *LineID);
DWORD GetControlID(DWORD LineID,const char *idStr,DWORD *ControlID);

int FAR PASCAL MCIString(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	char output[900];
	MCIERROR Err;
	if (Err = mciSendString(data,output,897,mWnd))
	{
		mciGetErrorString(Err,output,894);
		strcpy(data,"ERROR ");
	}
	else
		strcpy(data,"OK ");
	strcat(data,output);
	return 3;
}

#define MIXERTYPE_UNKNOWN	0
#define MIXERTYPE_BOOLEAN	1
#define MIXERTYPE_SIGNED	2
#define MIXERTYPE_UNSIGNED	3

int FAR PASCAL GetLineInfo(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	MIXERLINE MXLine;
	MXLine.cbStruct = sizeof(MIXERLINE);
	DWORD iLen = GetLineID(data,&MXLine.dwLineID);
	char * OutputModifier;

	if (iLen == 0)
		return 1;

	if (iLen == -1)
	{
		itoa(MXLine.dwLineID,data,10);
		return 3;
	}

	if (mixerGetLineInfo(0,&MXLine,MIXER_GETLINEINFOF_LINEID))
		return 1;

	OutputModifier = data+iLen;
	if (*OutputModifier == 0)		// Full name
	{
		strcpy(data,MXLine.szName);
		return 3;
	}

	OutputModifier++;
	if (!stricmp(OutputModifier,"sname"))	// Short name
	{
		strcpy(data,MXLine.szShortName);
		return 3;
	}
	else if (!stricmp(OutputModifier,"pname")) // Product name
	{
		if (MXLine.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED)
		{
			strcpy(data,MXLine.Target.szPname);
			return 3;
		}
		return 1;
	}
	else if (!stricmp(OutputModifier,"pid")) // Product ID
	{
		if (MXLine.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED)
		{
			itoa(MXLine.Target.wPid,data,10);
			return 3;
		}
		return 1;
	}
	else if (!stricmp(OutputModifier,"mid")) // Manufacturer ID
	{
		if (MXLine.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED)
		{
			itoa(MXLine.Target.wMid,data,10);
			return 3;
		}
		return 1;
	}
	else if (!stricmp(OutputModifier,"version")) // Driver version
	{
		if (MXLine.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED)
		{
			itoa(MXLine.Target.vDriverVersion,data,16);
			return 3;
		}
		return 1;
	}
	else if (!stricmp(OutputModifier,"dversion")) // Driver version in decimal
	{
		if (MXLine.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED)
		{
			itoa(MXLine.Target.vDriverVersion,data,10);
			return 3;
		}
		return 1;
	}
	else if (!stricmp(OutputModifier,"type")) // Line type
	{
		switch(MXLine.dwComponentType)
		{
#define TYPECHK(cst,str) case cst: strcpy(data,#str); break
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT,src waveout);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER,src midi);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC,src cd);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE,src mic);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_LINE,src line);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_DIGITAL,src digital);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER,src pcspeak);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY,src aux);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_ANALOG,src analog);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE,src tel);
			TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED,src undef);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,dst speakers);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_WAVEIN,dst wavein);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_DIGITAL,dst digital);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_HEADPHONES,dst headphones);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_LINE,dst line);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_MONITOR,dst monitor);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_TELEPHONE,dst tel);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_VOICEIN,dst voicein);
			TYPECHK(MIXERLINE_COMPONENTTYPE_DST_UNDEFINED,dst undef);
#undef TYPECHK
		default:
			strcpy(data,"unknown");
		}
		return 3;
	}
	else if (!stricmp(OutputModifier,"channels"))
	{
		itoa(MXLine.cChannels,data,10);
		return 3;
	}
	return 1;
}

#define CINFO_VALUE	0	// Control Info: Value
#define CINFO_NAME	1	// Control Info: Name
#define CINFO_SNAME	2	// Control Info: Short Name
#define CINFO_TYPE	3	// Control Info: Type
#define CINFO_CLASS	4	// Control Info: Class
#define CINFO_MIN	5	// Control Info: Minimum Value
#define CINFO_MAX	6	// Control Info: Maximum Value
#define CINFO_STEPS	7	// Control Info: Steps
#define CINFO_DTYPE	8	// Control Info: Data Type
int FAR PASCAL GetControlInfo(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	MIXERLINECONTROLS MLC;
	MIXERCONTROL MCtrl;
	
	DWORD DInfo = CINFO_VALUE;
	DWORD LineID;
	DWORD iLen = GetLineID(data,&LineID);

	if (iLen < 1)
		return 1;

	char *DPos = data + iLen;
	if (*DPos == 0)
		return 1;
	DPos++;

	MLC.cbStruct = sizeof(MIXERLINECONTROLS);
	MLC.cbmxctrl = sizeof(MIXERCONTROL);
	iLen = GetControlID(LineID,DPos,&MLC.dwControlID);

	if (iLen == -1)
	{		
		itoa(MLC.dwControlID,data,10);
		return 3;
	}

	if (iLen < 1)
		return 1;

	DPos += iLen;
	if (*DPos == 0)
		DInfo = CINFO_VALUE;
	else
	{
		DPos++;
		int SLen = strlen(DPos);
#define TYPECHK(cst,str) \
		if (!strnicmp(DPos,#str,SLen)) \
			DInfo = cst
		TYPECHK(CINFO_NAME,name);
		else TYPECHK(CINFO_SNAME,sname);
		else TYPECHK(CINFO_TYPE,type);
		else TYPECHK(CINFO_CLASS,class);
		else TYPECHK(CINFO_MIN,min);
		else TYPECHK(CINFO_MAX,max);
		else TYPECHK(CINFO_STEPS,steps);
		else TYPECHK(CINFO_DTYPE,dtype);
		else
			return 1;
#undef TYPECHK
	}


	MLC.cControls = 1;
	MLC.dwLineID = LineID;
	MLC.pamxctrl = &MCtrl;

	MMRESULT Err = mixerGetLineControls(0,&MLC,MIXER_GETLINECONTROLSF_ONEBYID);
	DWORD DType = ControlDataType(MCtrl.dwControlType);
	if (!Err)
	{
		switch(DInfo)
		{
		case CINFO_VALUE:
		{
			MIXERCONTROLDETAILS MXCDetail;
			MXCDetail.cbStruct = sizeof(MIXERCONTROLDETAILS);


			MXCDetail.dwControlID = MLC.dwControlID;
			MXCDetail.cChannels = 1;
			MXCDetail.cMultipleItems = MCtrl.cMultipleItems;

			switch(DType)	// Yes, they're all the same size. I don't care.
			{
			case MIXERTYPE_BOOLEAN:
				MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
				break;
			case MIXERTYPE_SIGNED:
				MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_SIGNED);
				break;
			case MIXERTYPE_UNSIGNED:
				MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
				break;
			default:
				return 1;
			}

			// Twice the size of a MIXERCONTROLDETAILS_LISTTEXT, just to be safe
			char SData[sizeof(MIXERCONTROLDETAILS_LISTTEXT)*2];	

			MXCDetail.paDetails = SData;

			if (mixerGetControlDetails(0,&MXCDetail,MIXER_GETCONTROLDETAILSF_VALUE))
				return 1;

			switch(DType)
			{
			case MIXERTYPE_BOOLEAN:
				strcpy(data,(((MIXERCONTROLDETAILS_BOOLEAN *)SData)->fValue ? "1" : "0"));
				break;
			case MIXERTYPE_SIGNED:
				itoa((signed int)((MIXERCONTROLDETAILS_SIGNED *)SData)->lValue,data,10);
				break;
			case MIXERTYPE_UNSIGNED:
				itoa((unsigned int)((MIXERCONTROLDETAILS_SIGNED *)SData)->lValue,data,10);
				break;
			}
			break;
		}
		case CINFO_NAME:
			strcpy(data,MCtrl.szName);
			break;
		case CINFO_SNAME:
			strcpy(data,MCtrl.szShortName);
			break;
		case CINFO_TYPE:
			switch(MCtrl.dwControlType)
			{
#define TYPECHK(cst,str) case cst: strcpy(data,#str); break;
				TYPECHK(MIXERCONTROL_CONTROLTYPE_CUSTOM,custom);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_BASS,bass);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_EQUALIZER,equalizer);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_FADER,fader);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_TREBLE,treble);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_VOLUME,volume);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MIXER,mixer);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MULTIPLESELECT,multi);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MUX,mux);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_SINGLESELECT,single);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_BOOLEANMETER,boolmeter);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_PEAKMETER,peakmeter);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_SIGNEDMETER,s-meter);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_UNSIGNEDMETER,u-meter);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_DECIBELS,decibels);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_PERCENT,percent);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_SIGNED,signed);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_UNSIGNED,unsigned);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_PAN,pan);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_QSOUNDPAN,qsoundpan);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_SLIDER,slider);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_BOOLEAN,boolean);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_BUTTON,button);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_LOUDNESS,loudness);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MONO,mono);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MUTE,mute);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_ONOFF,onoff);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_STEREOENH,stereoenh);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MICROTIME,microtime);
				TYPECHK(MIXERCONTROL_CONTROLTYPE_MILLITIME,millitime);
			default:
				strcpy(data,"unknown");
			}
			break;
		case CINFO_CLASS:
			switch(MCtrl.dwControlType & 0xF0000000)
			{
				TYPECHK(MIXERCONTROL_CT_CLASS_CUSTOM,_custom);
				TYPECHK(MIXERCONTROL_CT_CLASS_FADER,_fader);
				TYPECHK(MIXERCONTROL_CT_CLASS_LIST,_list);
				TYPECHK(MIXERCONTROL_CT_CLASS_METER,_meter);
				TYPECHK(MIXERCONTROL_CT_CLASS_NUMBER,_number);
				TYPECHK(MIXERCONTROL_CT_CLASS_SLIDER,_slider);
				TYPECHK(MIXERCONTROL_CT_CLASS_SWITCH,_switch);
				TYPECHK(MIXERCONTROL_CT_CLASS_TIME,_time);
			default:
				strcpy(data,"_unknown");
			}
			break;
#undef TYPECHK
		case CINFO_MIN:
			switch(DType)
			{
			case MIXERTYPE_BOOLEAN:
				strcpy(data,"0");
				break;
			case MIXERTYPE_SIGNED:
				itoa(MCtrl.Bounds.lMinimum,data,10);
				break;
			case MIXERTYPE_UNSIGNED:
				itoa(MCtrl.Bounds.dwMinimum,data,10);
				break;
			default:
				return 1;
			}
			break;
		case CINFO_MAX:
			switch(DType)
			{
			case MIXERTYPE_BOOLEAN:
				strcpy(data,"1");
				break;
			case MIXERTYPE_SIGNED:
				itoa(MCtrl.Bounds.lMaximum,data,10);
				break;
			case MIXERTYPE_UNSIGNED:
				itoa(MCtrl.Bounds.dwMaximum,data,10);
				break;
			default:
				return 1;
			}
			break;
		case CINFO_STEPS:
			itoa(MCtrl.Metrics.cSteps,data,10);
			break;
		case CINFO_DTYPE:
			switch(DType)
			{
			case MIXERTYPE_BOOLEAN:
				strcpy(data,"boolean");
				break;
			case MIXERTYPE_SIGNED:
				strcpy(data,"signed");
				break;
			case MIXERTYPE_UNSIGNED:
				strcpy(data,"unsigned");
				break;
			default:
				strcpy(data,"unknown");
			}
			break;
		default:
			return 1;
		}
		return 3;
	}
	return 1;
}

int FAR PASCAL SetControlValue(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	MIXERCONTROLDETAILS MXCDetail;
	MXCDetail.cbStruct = sizeof(MIXERCONTROLDETAILS);

	MIXERLINECONTROLS MLC;
	MIXERCONTROL MCtrl;
	MIXERLINE MXLine;
	
	DWORD DInfo = CINFO_NAME;
	DWORD LineID;
	DWORD iLen = GetLineID(data,&LineID);
	DWORD DType;

	if (iLen < 1)
		return 1;

	char *DPos = data + iLen;
	if (*DPos == 0)
		return 1;
	DPos++;
	
	MLC.cbStruct = sizeof(MIXERLINECONTROLS);
	MLC.cbmxctrl = sizeof(MIXERCONTROL);
	MLC.pamxctrl = &MCtrl;
	iLen = GetControlID(LineID,DPos,&MLC.dwControlID);

	if (iLen < 1)
		return 1;

	DPos += iLen;
	if (*DPos == 0)
		return 1;
	DPos++;

	MXLine.cbStruct = sizeof(MIXERLINE);
	MXLine.dwLineID = LineID;
	if (mixerGetLineInfo(0,&MXLine,MIXER_GETLINEINFOF_LINEID))
		return 1;
	if (mixerGetLineControls(0,&MLC,MIXER_GETLINECONTROLSF_ONEBYID))
		return 1;

	MXCDetail.dwControlID = MLC.dwControlID;
	MXCDetail.cChannels = 1;
	MXCDetail.cMultipleItems = MCtrl.cMultipleItems;

	DType = ControlDataType(MCtrl.dwControlType);
	char SData[sizeof(MIXERCONTROLDETAILS_LISTTEXT)*2];	
	MXCDetail.paDetails = SData;

	switch(DType)	// Yes, they're all the same size. I don't care.
	{
	case MIXERTYPE_BOOLEAN:
		MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
		((MIXERCONTROLDETAILS_BOOLEAN *)SData)->fValue = (atoi(DPos) ? 1 : 0);
		break;
	case MIXERTYPE_SIGNED:
		MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_SIGNED);
		((MIXERCONTROLDETAILS_SIGNED *)SData)->lValue = atoi(DPos);
		break;
	case MIXERTYPE_UNSIGNED:
		if (*DPos == '-')
			return 1;
		((MIXERCONTROLDETAILS_SIGNED *)SData)->lValue = atoi(DPos);
		MXCDetail.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
		break;
	default:
		return 1;
	}

	if (mixerSetControlDetails(0,&MXCDetail,MIXER_SETCONTROLDETAILSF_VALUE))
		return 1;

	strcpy(data,"OK");
	return 3;
}

// Gets the data type of a mixer control from its control type
DWORD ControlDataType(DWORD MixerType)
{
	switch(MixerType)
	{
	case MIXERCONTROL_CONTROLTYPE_BOOLEANMETER:
	case MIXERCONTROL_CONTROLTYPE_BOOLEAN:
	case MIXERCONTROL_CONTROLTYPE_BUTTON:
	case MIXERCONTROL_CONTROLTYPE_LOUDNESS:
	case MIXERCONTROL_CONTROLTYPE_MONO:
	case MIXERCONTROL_CONTROLTYPE_MUTE:
	case MIXERCONTROL_CONTROLTYPE_ONOFF:
	case MIXERCONTROL_CONTROLTYPE_STEREOENH:
	case MIXERCONTROL_CONTROLTYPE_MIXER:
	case MIXERCONTROL_CONTROLTYPE_MULTIPLESELECT:
	case MIXERCONTROL_CONTROLTYPE_MUX:
	case MIXERCONTROL_CONTROLTYPE_SINGLESELECT:
		return MIXERTYPE_BOOLEAN;
	case MIXERCONTROL_CONTROLTYPE_UNSIGNEDMETER:
	case MIXERCONTROL_CONTROLTYPE_UNSIGNED:
	case MIXERCONTROL_CONTROLTYPE_BASS:
	case MIXERCONTROL_CONTROLTYPE_EQUALIZER:
	case MIXERCONTROL_CONTROLTYPE_FADER:
	case MIXERCONTROL_CONTROLTYPE_TREBLE:
	case MIXERCONTROL_CONTROLTYPE_VOLUME:
	case MIXERCONTROL_CONTROLTYPE_MICROTIME:
	case MIXERCONTROL_CONTROLTYPE_MILLITIME:
	case MIXERCONTROL_CONTROLTYPE_PERCENT:
		return MIXERTYPE_UNSIGNED;
	case MIXERCONTROL_CONTROLTYPE_PEAKMETER:
	case MIXERCONTROL_CONTROLTYPE_SIGNEDMETER:
	case MIXERCONTROL_CONTROLTYPE_SIGNED:
	case MIXERCONTROL_CONTROLTYPE_DECIBELS:
	case MIXERCONTROL_CONTROLTYPE_PAN:
	case MIXERCONTROL_CONTROLTYPE_QSOUNDPAN:
	case MIXERCONTROL_CONTROLTYPE_SLIDER:
		return MIXERTYPE_SIGNED;
	default:
		return MIXERTYPE_UNKNOWN;
	}
}

// Gets a line ID in one of a few forms
// src [analog|aux|cd|digital|line|mic|pcspeak|midi|tel|undef|waveout] <num>
// dst [digital|headphones|line|monitor|speakers|tel|undef|voicein|wavein] <num>
// Use 0 as <num> for a total
// Return value is string length parsed on success
// 0 on failure
// -1 on line count
DWORD GetLineID(const char *idStr,DWORD *LineID)
{
	MIXERLINE IDInfo;
	IDInfo.cbStruct = sizeof(MIXERLINE);
	DWORD GetType;
	DWORD LookForType = -1;
	DWORD TypeLen;
	DWORD idNum;
	char *TypeName;
	char *NumLoc;
	if (!strnicmp(idStr,"src ",4))
		GetType = MIXER_GETLINEINFOF_SOURCE;
	else if (!strnicmp(idStr,"dst ",4))
		GetType = MIXER_GETLINEINFOF_DESTINATION;
	else
		return 0;
	TypeName=(char *)((long)idStr+4);
	NumLoc=strchr(TypeName,' ');
	if (NumLoc != NULL)
		TypeLen = NumLoc-TypeName;
	else
		TypeLen = strlen(TypeName);
#define TYPECHK(cst,txt) \
	if (!strnicmp(TypeName,#txt,TypeLen)) \
		LookForType = cst
	if (GetType == MIXER_GETLINEINFOF_SOURCE)
	{
		// This isn't alphabetical order, but I figure the most-used types should come first
		TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT,waveout);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER,midi);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC,cd);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE,mic);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_LINE,line);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_DIGITAL,digital);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER,pcspeak);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY,aux);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_ANALOG,analog);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE,tel);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED,undef);
		else if ((TypeName[0] < '0') || (TypeName[0] > '9'))
			return 0;
	}
	else
	{
		TYPECHK(MIXERLINE_COMPONENTTYPE_DST_SPEAKERS,speakers);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_WAVEIN,wavein);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_DIGITAL,digital);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_HEADPHONES,headphones);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_LINE,line);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_MONITOR,monitor);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_TELEPHONE,tel);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_VOICEIN,voicein);
		else TYPECHK(MIXERLINE_COMPONENTTYPE_DST_UNDEFINED,undef);
		else if ((TypeName[0] < '0') || (TypeName[0] > '9'))
			return 0;
	}
	if ((NumLoc == NULL) || (LookForType == -1))
		NumLoc = TypeName-1;
	if (NumLoc != NULL)
	{
		NumLoc++;
		idNum = atoi(NumLoc);
	}
	else
		return 0;

	if ((LookForType != -1) || (idNum == 0))
	{
		// Now we have to scan through the IDs since we care what they are

		DWORD idCount = 0;
		DWORD idCur = 0;
		IDInfo.dwSource = 0;
		IDInfo.dwDestination = 0;
		while(!mixerGetLineInfo(0,&IDInfo,GetType))
		{
			if ((IDInfo.dwComponentType == LookForType) || ((idNum == 0) && (LookForType == -1)))
			{
				idCount++;
				if (idCount == idNum)
				{
					*LineID = IDInfo.dwLineID;
					char * tmp = strchr(NumLoc,' ');
					if (tmp == NULL)
						return strlen(idStr);
					else
						return tmp-idStr;
				}
			}
			idCur++;
			if (GetType == MIXER_GETLINEINFOF_SOURCE)
			{
				IDInfo.dwSource = idCur;
				IDInfo.dwDestination = 0;
			}
			else
			{
				IDInfo.dwSource = 0;
				IDInfo.dwDestination = idCur;
			}
		}
		if (idNum == 0)
		{
			*LineID = idCount;
			return -1;
		}
		return 0;
	}
	else
	{
		// No, we don't. Maybe.
		if (GetType == MIXER_GETLINEINFOF_SOURCE)
		{
			IDInfo.dwSource = idNum - 1;
			IDInfo.dwDestination = 0;
		}
		else
		{
			IDInfo.dwSource = 0;
			IDInfo.dwDestination = idNum - 1;
		}
		if (mixerGetLineInfo(0,&IDInfo,GetType))
			return 0;
		*LineID = IDInfo.dwLineID;
		char *tmp = strchr(NumLoc,' ');
		if (tmp == NULL)
			return strlen(idStr);
		else
			return tmp-idStr;
	}

}


// When given a LineID, gets the ID of a control off of idStr, which is in the form:
// [] <num>
// Use 0 for <num> for a total
// Return value is string length parsed on success
// 0 on failure
// -1 on control count
#define MAX_CONTROLS 16				// Can handle up to 16 controls.
									// I don't want to adjust this dynamically,
									// so I can avoid memory leaks -- a lot of malloc()'s
									// is never a good thing.

DWORD GetControlID(DWORD LineID,const char *idStr,DWORD *ControlID)
{
	MIXERLINECONTROLS MLC;
	MIXERCONTROL MCtrl[MAX_CONTROLS];
	DWORD WantID;
	DWORD LookForType = -1;
	bool fClass = 0;		// Are we looking for a class?
	
	MIXERLINE MXLine;
	MXLine.cbStruct = sizeof(MIXERLINE);

	MXLine.dwLineID = LineID;

	if (mixerGetLineInfo(0,&MXLine,MIXER_GETLINEINFOF_LINEID))	// Wasn't a valid line ID
		return 0;

	MLC.cbStruct = sizeof(MIXERLINECONTROLS);
	MLC.cbmxctrl = sizeof(MIXERCONTROL);
	MLC.cControls = MXLine.cControls;
	MLC.dwLineID = MXLine.dwLineID;
	MLC.pamxctrl = MCtrl;

	const char *TypeName = idStr;
	DWORD TypeLen;
	const char *NumLoc = strchr(idStr,' ');
	if (NumLoc == NULL)
		TypeLen = strlen(idStr);
	else
		TypeLen = NumLoc - idStr;
	
	TYPECHK(MIXERCONTROL_CT_CLASS_CUSTOM,_custom);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_CUSTOM,custom);
	else TYPECHK(MIXERCONTROL_CT_CLASS_FADER,_fader);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_BASS,bass);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_EQUALIZER,equalizer);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_FADER,fader);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_TREBLE,treble);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_VOLUME,volume);
	else TYPECHK(MIXERCONTROL_CT_CLASS_LIST,_list);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MIXER,mixer);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MULTIPLESELECT,multi);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MUX,mux);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_SINGLESELECT,single);
	else TYPECHK(MIXERCONTROL_CT_CLASS_METER,_meter);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_BOOLEANMETER,boolmeter);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_PEAKMETER,peakmeter);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_SIGNEDMETER,s-meter);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_UNSIGNEDMETER,u-meter);
	else TYPECHK(MIXERCONTROL_CT_CLASS_NUMBER,_number);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_DECIBELS,decibels);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_PERCENT,percent);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_SIGNED,signed);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_UNSIGNED,unsigned);
	else TYPECHK(MIXERCONTROL_CT_CLASS_SLIDER,_slider);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_PAN,pan);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_QSOUNDPAN,qsoundpan);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_SLIDER,slider);
	else TYPECHK(MIXERCONTROL_CT_CLASS_SWITCH,_switch);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_BOOLEAN,boolean);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_BUTTON,button);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_LOUDNESS,loudness);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MONO,mono);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MUTE,mute);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_ONOFF,onoff);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_STEREOENH,stereoenh);
	else TYPECHK(MIXERCONTROL_CT_CLASS_TIME,_time);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MICROTIME,microtime);
	else TYPECHK(MIXERCONTROL_CONTROLTYPE_MILLITIME,millitime);
	else if ((TypeName[0] < '0') || (TypeName[0] > '9'))
		return 0;

	if (TypeName[0] == '_')
		fClass = 1;

	if ((NumLoc == NULL) || (LookForType == -1))
		NumLoc = idStr-1;
	if (NumLoc != NULL)
		NumLoc++;
	else
		return 0;

	if ((NumLoc[0] < '0') || (NumLoc[0] > '9'))
		return 0;

	WantID = atoi(NumLoc);

	if ((WantID == 0) && (LookForType == -1))
	{
		*ControlID = MLC.cControls;
		return -1;
	}


	if ((WantID > MXLine.cControls) || (MXLine.cControls > MAX_CONTROLS))
		return 0;

	if(mixerGetLineControls(0,&MLC,MIXER_GETLINECONTROLSF_ALL))
		return 0;

	if ((LookForType != -1) || (WantID == 0))
	{
		DWORD IDScan;
		DWORD Count=0;
		bool Found = 0;
		for(IDScan=0;IDScan < MLC.cControls;IDScan++)
		{
			if ((fClass && ((MCtrl[IDScan].dwControlType & 0xF0000000) == LookForType))
				|| (MCtrl[IDScan].dwControlType == LookForType))
			{
				Count++;
				if (Count == WantID)
				{
					WantID = IDScan+1;
					Found = 1;
					break;
				}
			}
		}
		if (WantID == 0)
		{
			*ControlID = Count;
			return -1;
		}
		if (!Found)
			return 0;
	}

	*ControlID = MCtrl[WantID-1].dwControlID;

	char *tmp = strchr(NumLoc,' ');
	if (tmp == NULL)
		return strlen(idStr);
	else
		return tmp-idStr;
#undef TYPECHK
}

int FAR PASCAL DLLInfo(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause)
{
	strcpy(data,AUTHINFO);
	return 3;
}
#ifdef __cplusplus
}
#endif